from fastapi import Depends, FastAPI, Request
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from fastapi.middleware.cors import CORSMiddleware

from chlore.request import request_context
from chlore.logging import setup_app_logging, setup_standard_logging
from chlore.config import CONFIG
from chlore.errors import ChloreError
from chlore.etna_auth import header_auth, any_auth


app = FastAPI(dependencies=[Depends(any_auth)])


@app.on_event("startup")
def setup_logging():
    config = CONFIG.logging
    setup_standard_logging(config)
    setup_app_logging(config)


@app.on_event("startup")
def setup_cors():
    if CONFIG.cors is not None and CONFIG.cors.allowed_origin_re:
        app.add_middleware(
            CORSMiddleware,
            allow_origin_regex=CONFIG.cors.allowed_origin_re,
            allow_credentials=True,
            allow_methods=["GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS", "HEAD"],
            allow_headers=["Content-Type", "turbo-frame", "Cookie"],
            max_age=600,
        )


@app.middleware("http")
async def setup_request_context(request, call_next):
    with request_context(request):
        return await call_next(request)


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request, exc: RequestValidationError):
    return JSONResponse(status_code=400, content={"detail": exc.errors()})


@app.exception_handler(ChloreError)
async def exception_handler_hub(request: Request, exc: ChloreError):
    return JSONResponse(
        status_code=exc.status_code,
        content={"message": exc.detail, "error_code": exc.error_code},
    )



from contextlib import contextmanager
from typing import cast

from fastapi.testclient import TestClient
import requests

from .dependencies import override_dependency


@contextmanager
def capture_jinja_context(app, jinja_mixin):
    capture_context = {}
    original_render = jinja_mixin.render

    def render(self, template: str, context: dict, **kwargs) -> dict:
        capture_context.update(context)
        return original_render(self, template, context=context, **kwargs)

    try:
        jinja_mixin.render = render
        yield capture_context
    finally:
        jinja_mixin = original_render


class WebTestResponse(requests.Response):
    def __init__(self, req: requests.Response, jinja_context: dict):
        super().__init__()
        for k, v in req.__dict__.items():
            self.__dict__[k] = v
        self.jinja_context = jinja_context


class WebTestClient(TestClient):
    """
    A TestClient subclass specialized to test web applications
    """

    def __init__(self, *args, jinja_mixin, **kwargs):
        super().__init__(*args, **kwargs)
        self.jinja_mixin = jinja_mixin

    @staticmethod
    def _add_default_kwargs(kw):
        kw.setdefault("allow_redirects", False)
        return kw

    def request(self, *args, **kwargs) -> WebTestResponse:
        kwargs = self._add_default_kwargs(kwargs)
        with capture_jinja_context(self.app, self.jinja_mixin) as context:
            resp = super().request(*args, **kwargs)
            return WebTestResponse(resp, context)

    def get(self, *args, **kwargs) -> WebTestResponse:
        kwargs = self._add_default_kwargs(kwargs)
        return cast(WebTestResponse, super().get(*args, **kwargs))

    def post(self, *args, **kwargs) -> WebTestResponse:
        kwargs = self._add_default_kwargs(kwargs)
        return cast(WebTestResponse, super().post(*args, **kwargs))

    def put(self, *args, **kwargs) -> WebTestResponse:
        kwargs = self._add_default_kwargs(kwargs)
        return cast(WebTestResponse, super().put(*args, **kwargs))

    def delete(self, *args, **kwargs) -> WebTestResponse:
        kwargs = self._add_default_kwargs(kwargs)
        return cast(WebTestResponse, super().put(*args, **kwargs))



import sqlalchemy
from sqlalchemy.orm import sessionmaker, Session, declarative_base
from sqlalchemy.sql.elements import BinaryExpression
from functools import cache
from fastapi import Depends

from chlore.config import CONFIG


Base = declarative_base()
metadata = Base.metadata


class DatabaseSession(Session):
    pass


@cache
def get_engine():
    if CONFIG.database is None:
        raise ConfigurationError("Chlore: SQL Engine cannot be used without a config entry.")
    return sqlalchemy.create_engine(CONFIG.database.url, pool_pre_ping=True)


@cache
def get_session_maker():
    engine = get_engine()
    return sessionmaker(engine, class_=DatabaseSession)


def session() -> DatabaseSession:
    """
    FastAPI dependency to acquire the database handle. For elegance, one should
    not import this directly into it's namespace, but rather import the module
    and write `Depends(database.session)`.
    """
    session_maker = get_session_maker()
    with session_maker() as session:
        yield session



import json
from contextlib import contextmanager
from dataclasses import dataclass

import pika
from pika.adapters.blocking_connection import BlockingChannel, BlockingConnection

from .config import CONFIG


@dataclass
class Publisher:
    """
    Pre-configured publisher

    ```
    @contextmanager
    def student_notice() -> Publisher:
        with rabbitmq_channel() as channel:
            yield Publisher(channel=channel, exchange="etna", routing_key="student_notice")


    with student_notice() as notice:
        notice.publish(my_bytes)
    ```
    """

    channel: BlockingChannel
    exchange: str
    routing_key: str

    def publish(self, body: bytes, properties=None, mandatory: bool = False):
        self.channel.basic_publish(self.exchange, self.routing_key, body, properties=properties, mandatory=mandatory)


class JSONPublisher(Publisher):
    """
    Pre-configured JSON publisher

    ```
    @contextmanager
    def student_notice() -> Publisher:
        with rabbitmq_channel() as channel:
            yield JSONPublisher(channel=channel, exchange="etna", routing_key="student_notice")


    with student_notice() as notice:
        notice.publish(my_data)
    ```
    """

    def publish(self, body, properties=None, mandatory: bool = False):
        encoded_json_body = json.dumps(body).encode()
        return super().publish(encoded_json_body, properties=properties, mandatory=mandatory)


@contextmanager
def rabbitmq_connection() -> BlockingConnection:
    params = pika.URLParameters(CONFIG.rabbitmq.url)
    with pika.BlockingConnection(params) as connection:
        yield connection


@contextmanager
def rabbitmq_channel() -> BlockingChannel:
    with rabbitmq_connection() as connection:
        with connection.channel() as channel:
            yield channel



import typing
from fastapi import Request
from enum import IntEnum
from dataclasses import dataclass


class FlashType(IntEnum):
    DEBUG = 0
    INFO = 1
    SUCCESS = 2
    WARNING = 3
    ERROR = 4


@dataclass
class UiFlashMessage:
    "Represent a flash message, with the purpose of rendering in jinja"

    message: str
    category: FlashType

    @property
    def bootstrap_classes(self):
        match self.category:
            case FlashType.SUCCESS:
                return "bg-success bg-opacity-50"
            case FlashType.ERROR:
                return "bg-danger bg-opacity-50"
            case FlashType.WARNING:
                return "bg-warning bg-opacity-50"
            case _:
                return ""


def add_message(request: Request, category: FlashType, message: str) -> None:
    if "_messages" not in request.session:
        request.session["_messages"] = []
    request.session["_messages"].append({"message": message, "category": category})


def get_messages(request: Request):
    return request.session.pop("_messages") if "_messages" in request.session else []


class FlashMessages:
    def __init__(self, request: Request):
        self.request = request

    def add(self, category: FlashType, message: str):
        return add_message(self.request, category, message)

    def pop_all(self):
        return [UiFlashMessage(**m) for m in get_messages(self.request)]

    def debug(self, message: str):
        self.add(FlashType.DEBUG, message)

    def info(self, message: str):
        self.add(FlashType.INFO, message)

    def success(self, message: str):
        self.add(FlashType.SUCCESS, message)

    def warning(self, message: str):
        self.add(FlashType.WARNING, message)

    def error(self, message: str):
        self.add(FlashType.ERROR, message)

    @classmethod
    def __call__(cls, request: Request):
        "Small hack: make this Class a fastapi dependable."
        return FlashMessages(request)



from contextlib import contextmanager
from contextvars import ContextVar
import uuid

from fastapi import Request

_REQUEST: ContextVar[Request] = ContextVar("_CURRENT_REQUEST")


def _set_request(request: Request):
    request.state.id = str(uuid.uuid4())
    return _REQUEST.set(request)


def _reset_request(token):
    _REQUEST.reset(token)


def get_request() -> Request:
    """
    Retrieve the current request
    """
    return _REQUEST.get()


@contextmanager
def request_context(request):
    ctx_token = _set_request(request)
    try:
        yield
    finally:
        _reset_request(ctx_token)



class ChloreError(Exception):
    status_code = 500
    error_code = "EUNKNOWN"
    detail = "Erreur inconnue."


class InternalError(ChloreError):
    """Base class for coding/config error.
    Raise this if the application state is incorrect due to a programmer's error, a misconfiguration, etc...
    """
    status_code = 500
    error_code = "EINTERNAL"
    detail = "Internal error."


class ConfigurationError(InternalError):
    status_code = 500
    detail = "Internal error: application configuration problem."


class ExternalApiError(ChloreError):
    status_code = 500
    error_code = "EUNKNOWN"
    detail = "Unexpected error when using external API."


class NotFound(ChloreError):
    status_code = 404
    error_code = "ENOENT"
    detail = "No such resource."

    
class PermissionDenied(ChloreError):
    status_code = 403
    error_code = "EPERM"
    detail = "You do not have the required permissions."



import base64
from contextlib import contextmanager
from functools import cache
import json
from pathlib import Path
import urllib.parse

from fastapi import FastAPI, Cookie, Depends, Header, HTTPException, params
from fastapi.requests import Request
import OpenSSL
import pydantic
import structlog
import requests

from chlore.config import CONFIG
from chlore.logging import request_logger

logger = structlog.get_logger("chlore.auth")


class InvalidToken(Exception):
    pass


def _parse_and_verify_etna_token(raw_token: str, cert: OpenSSL.crypto.X509 | None) -> dict:
    try:
        decoded = json.loads(base64.b64decode(raw_token))
        identity = decoded["identity"]
        signature = decoded["signature"]
    except Exception as e:
        raise InvalidToken("cannot parse etna auth token") from e

    try:
        if cert is not None:
            OpenSSL.crypto.verify(cert, base64.b64decode(signature), identity.encode(), "sha1")
    except Exception as e:
        raise InvalidToken("etna auth token verification failed") from e

    try:
        return json.loads(base64.b64decode(identity))
    except Exception as e:
        raise InvalidToken("cannot parse etna auth identity") from e


@cache
def _load_auth_key():
    """
    Load the public key from Config.auth.pubkey_path into a X509 signing certificate.
    """

    path = Path(CONFIG.auth.pubkey_path)
    logger.info("Loading auth signing certificates", path=str(path))
    if path == "/dev/null":
        return None
    content = path.read_text()
    pubkey = OpenSSL.crypto.load_publickey(OpenSSL.crypto.FILETYPE_PEM, content)
    cert = OpenSSL.crypto.X509()
    cert.set_pubkey(pubkey)
    return cert


def _extract_cookie_token(authenticator: str = Cookie(default=None)) -> str | None:
    if authenticator is not None:
        return urllib.parse.unquote_plus(authenticator)
    return None


def _extract_header_token(authorization: str = Header(default=None)) -> str | None:
    if authorization is not None:
        token = authorization.split("Legacy ")[-1]
        if token and len(token) != len(authorization):
            return token
    return None


def _extract_any_token(
    cookie_token: str = Depends(_extract_cookie_token),
    header_token: str = Depends(_extract_header_token),
) -> str | None:
    return cookie_token or header_token or None


def _add_identity(request: Request, token: str | None):
    if token is not None:
        cert = _load_auth_key()
        try:
            request.state.identity = _parse_and_verify_etna_token(token, cert)
            request.state.token = token
        except InvalidToken as e:
            request_logger().exception("cannot parse legacy auth token")
            # We want a 400 (validation error) instead of a 401 here. This avoid
            # a redirection loop to the login page, and provides the user with a nice
            # explaination why that request cannot be fullfiled. As this scenario (an
            # invalid token) is more likely to arise in dev, this seems to be the most
            # sensible status.
            raise HTTPException(400, detail="invalid auth token") from e


def any_auth(request: Request, token: str | None = Depends(_extract_any_token)):
    _add_identity(request, token)


def cookie_auth(request: Request, token: str | None = Depends(_extract_cookie_token)):
    _add_identity(request, token)


def header_auth(request: Request, token: str | None = Depends(_extract_header_token)):
    _add_identity(request, token)


class Logas(pydantic.BaseModel):
    login: str
    id: int
    roles: list[str]
    email: str
    login_date: str

    @pydantic.root_validator(pre=True)
    def _rename_groups_to_roles(cls, values):
        values["roles"] = values.get("groups", [])
        return values


class Identity(pydantic.BaseModel):
    login: str
    id: int
    roles: list[str]
    email: str
    login_date: str
    logas: bool | Logas

    @pydantic.root_validator(pre=True)
    def _rename_groups_to_roles(cls, values):
        values["roles"] = values.get("groups", [])
        return values

    def has_role(self, role):
        return role in self.roles

    def ensure_has_role(self, role):
        "Raise HTTPException(403) unless this identity has given role"
        if not self.has_role(role):
            raise HTTPException(403, detail="permission denied")


def identity(request: Request) -> Identity:
    identity = getattr(request.state, "identity", None)
    if identity is None:
        raise HTTPException(401, detail="unauthorized")
    return Identity(**identity)


def _request_token(request: Request) -> str:
    token = getattr(request.state, "token", None)
    if token is None:
        raise HTTPException(401, detail="unauthorized")
    return token


class RolesAllowed:
    def __init__(self, *roles: str):
        self.roles_allowed = roles

    def __call__(self, identity: Identity = Depends(identity)):
        if any(identity.has_role(role) for role in self.roles_allowed):
            return self.roles_allowed
        raise HTTPException(403, detail="permission denied")


requires_adm = RolesAllowed("adm")


@contextmanager
def fake_identity(app: FastAPI, wanted_identity: Identity):
    try:
        app.dependency_overrides[identity] = lambda: wanted_identity
        app.dependency_overrides[EtnaToken.from_request] = lambda: ""
        yield wanted_identity
    finally:
        if identity in app.dependency_overrides:
            del app.dependency_overrides[identity]
        if EtnaToken.from_request in app.dependency_overrides:
            del app.dependency_overrides[EtnaToken.from_request]


class EtnaToken:
    """
    Abstraction over ETNA authentication information needed to use Etna APIs and services.

    It can be constructed from multiple sources, such as an HTTP request or a raw base64-encoded token.
    """

    def __init__(self, value: str):
        self.value = value

    @property
    def as_base64(self) -> str:
        return self.value

    @classmethod
    def from_base64(cls, token: str) -> "EtnaToken":
        """
        Construct an EtnaToken from a raw base64-encoded token
        """
        return cls(token)

    @classmethod
    def from_request(cls, request: Request) -> "EtnaToken":
        """
        Construct an EtnaToken from a given request (or raise an HTTPException(401) if the request is unauthenticated)
        """
        return cls(_request_token(request))

    @classmethod
    def from_login_and_password(cls, login, password):
        try:
            auth_url = CONFIG.auth.url.rstrip("/") + "/identity"
            login_form = dict(login=login, password=password)
            response = requests.post(auth_url, data=login_form)
            if response.status_code == 200:
                return cls.from_base64(response.cookies["authenticator"])
            else:
                logger.warning(f"auth login failed", url=auth_url, login=login, status=response.status_code)
        except Exception as E:
            logger.exception(f"failed to login against legacy auth", url=auth_url, login=login)
            raise
        return cls()  # will raise an exception as there is no credential provided.



from contextlib import contextmanager
from typing import Callable


@contextmanager
def override_dependency(app, dependency: Callable, replacement: Callable):
    """
    Override a given dependency in a FastAPI app by replacing it with another function
    """
    previous = app.dependency_overrides.get(dependency)
    app.dependency_overrides[dependency] = replacement
    try:
        yield
    finally:
        if previous is not None:
            app.dependency_overrides[dependency] = previous
        else:
            app.dependency_overrides.pop(dependency)



from fastapi import FastAPI, Depends
from fastapi.testclient import TestClient

from chlore.etna_auth import Identity, identity, any_auth, requires_adm, fake_identity


def test_auth():
    app = FastAPI()
    client = TestClient(app)

    @app.get("/any_auth")
    def any_auth_endpoint(identity: dict = Depends(identity)):
        return "OK"

    @app.get("/adm_only", dependencies=[Depends(any_auth), Depends(requires_adm)])
    def adm_only_endpoint():
        return "OK"

    r = client.get("/any_auth")
    assert r.status_code == 401
    ident1 = Identity(
        login="student_s", id=1, groups=["student"], email="student_s@staging.etna.io", login_date="", logas=False
    )
    ident2 = Identity(login="adm_a", id=2, groups=["adm"], email="adm_a@staging.etna.io", login_date="", logas=False)
    with fake_identity(app, ident1):
        r = client.get("/any_auth")
        assert r.status_code == 200
        r = client.get("/adm_only")
        assert r.status_code == 403
    r = client.get("/any_auth")
    assert r.status_code == 401



import inspect
from typing import ClassVar, Type, get_origin, get_type_hints

from fastapi import APIRouter, Depends
import structlog

from .logging import request_logger


def _patch_class(cls: Type):
    """
    Patch a class P defined as:
    ```
    class P:
        x: int = Depends(y)
    ```
    To turn it into a class defined like:
    ```
    class P:
        def __init__(self, *, x: int = Depends(y)):
            self.x = x
    ```
    """

    def to_new_param(name: str, type_hint) -> inspect.Parameter:
        return inspect.Parameter(
            name=name,
            kind=inspect.Parameter.KEYWORD_ONLY,  # use kwargs in order to access them by name later
            annotation=type_hint,
            default=getattr(cls, name, Ellipsis),
        )

    def new_init(self, *_, **kwargs):
        for p in new_param_names:
            setattr(self, p, kwargs.get(p))

    original_signature = inspect.signature(cls.__init__)
    new_params = [to_new_param(n, h) for n, h in get_type_hints(cls).items() if get_origin(h) is not ClassVar]
    cls.__signature__ = original_signature.replace(parameters=new_params)
    new_param_names = [p.name for p in new_params]
    cls.__init__ = new_init


def _patch_endpoint(cls: Type, endpoint):
    """
    Patch a FastAPI endpoint defined as a method of a given class:
    ```
    @app.get("/{item}")
    def index(self, item: int):
        pass
    ```
    To turn it into a method defined like:
    ```
    @app.get("/{item}")
    def index(self = Depends(cls), *, item: int):
        pass
    ```
    """
    original_signature = inspect.signature(endpoint)
    original_params = iter(original_signature.parameters.values())
    self_param = next(original_params).replace(default=Depends(cls))
    kwargs = [p.replace(kind=inspect.Parameter.KEYWORD_ONLY) for p in original_params]
    endpoint.__signature__ = original_signature.replace(parameters=[self_param] + kwargs)


class Controller:
    """
    Base class allowing to write FastAPI endpoints as methods and specify class-level dependencies

    Example:
    ```
    class MyClass(Controller, router=router):
        some_dependency: str = Depends(get_some_dependency)

        @router.get("/index")
        def index(self):
            return "OK " + self.some_dependency
    ```
    """

    def __init_subclass__(cls, router: APIRouter = None, **kwargs):
        if router is not None:
            _patch_class(cls)

            methods = {v for _, v in inspect.getmembers(cls, inspect.isfunction)}
            for route in router.routes:
                if route.endpoint not in methods:
                    continue
                _patch_endpoint(cls, route.endpoint)


class LoggerMixin:
    """
    A controller mixin to add a dependency over the current request logger
    """

    logger: structlog.stdlib.BoundLogger = Depends(request_logger)



import os
import re
import pytest
import requests_mock
import requests

from sqlalchemy.engine import Engine, create_engine
from sqlalchemy.orm import Session
from sqlalchemy import text
from fastapi.testclient import TestClient
from requests_mock import ANY

import chlore.database
from chlore.config import CONFIG
from chlore.app import app
from chlore.testing.dependencies import override_dependency


@pytest.fixture(scope="session")
def engine() -> Engine:
    "SQL Alchemy engine, configured using TEST_DATABASE_* env vars"
    if CONFIG.database is None:
        return None
    else:
        host = os.environ["TEST_DATABASE_HOST"]
        user = os.environ["TEST_DATABASE_USER"]
        password = os.environ["TEST_DATABASE_PASS"]
        return create_engine(f"mysql+pymysql://{user}:{password}@{host}/my_sap", pool_pre_ping=True)


@pytest.fixture
def session(engine):
    "SQL Alchemy session, tuned to be the same instance used by fastapi's dependency."
    connection = engine.connect()
    transaction = connection.begin()
    savepoint = connection.begin_nested()
    session = Session(bind=connection, expire_on_commit=False, join_transaction_mode="create_savepoint")
    with override_dependency(app, chlore.database.session, lambda: session):
        yield session
    session.close()
    transaction.rollback()
    connection.close()


def truncate_tables(engine):
    with Session(engine, expire_on_commit=False) as session:
        session.execute(text("SET FOREIGN_KEY_CHECKS = 0;"))
        for table in chlore.database.metadata.tables:
            session.execute(text(f"TRUNCATE TABLE `{table}`;"))
        session.execute(text("SET FOREIGN_KEY_CHECKS = 1;"))
        session.commit()


@pytest.fixture(scope="session", autouse=True)
def fresh_database(engine):
    if CONFIG.database is None:
        yield None
    else:
        chlore.database.metadata.create_all(engine)
        yield
        truncate_tables(engine)
    

@pytest.fixture(scope="module")
def test_client():
    "A fastapi.TestClient instance for this app."
    with TestClient(app) as client:
        yield client


def _url_for(base, **params):
    """Contruct urls with query arguments the same way requests does.
    This is a workaround needed to mock requests with arguments, as the code in
    requests_mock expects the query argument in the url to be in the exact same
    order both in the requests and in the mock definition.
    """
    request = requests.Request("GET", base, params=params).prepare()
    return request.url


@pytest.fixture
def http_mock(test_client, requests_mock):
    "requests_mock.Mocker instance tuned to not interfere with fastapi.TestClient."
    # we need to whitelist the fastapi TestClient's base url.
    base_url_re = re.escape(test_client.base_url)
    matcher = re.compile(rf"{base_url_re}(/.*)?")
    requests_mock.register_uri(ANY, matcher, real_http=True)
    requests_mock.url_for = _url_for
    return requests_mock



import os
from contextlib import contextmanager

from pydantic import BaseSettings

from chlore.config import CONFIG, DatabaseConfig, from_env


@contextmanager
def push_env(values: dict):
    old_env = os.environ.copy()
    for k, v in values.items():
        os.environ[k] = v
    yield os.environ
    os.environ = old_env


def test_from_env_without_prefix():
    with push_env({"URL": "some://url"}):
        database = from_env(DatabaseConfig)
        assert database.url == "some://url"


def test_from_env_with_prefix():
    with push_env({"DATABASE_URL": "some://url"}):
        database = from_env(DatabaseConfig, with_prefix="DATABASE_")
        assert database.url == "some://url"


class OtherConfig(BaseSettings):
    url: str


def test_from_env_registration():
    with push_env({"URL": "some://url"}):
        other = from_env(OtherConfig)
        assert other.url == "some://url"

    with push_env({"URL": "some://url"}):
        database = from_env(DatabaseConfig)
        assert database.url == "some://url"
        assert CONFIG.database is database



import inspect
from typing import Callable, ClassVar, Type, TypeVar

from fastapi import APIRouter, Depends, Request
from wtforms import Form

from .controller import Controller


async def _extract_form_data(request: Request):
    return await request.form()


WholeForm = lambda: Depends(_extract_form_data)


def _copy_dependencies(source_function: Callable, dest_function: Callable):
    source_signature = inspect.signature(source_function)
    dest_signature = inspect.signature(dest_function)
    _, _, *params_to_copy = source_signature.parameters.values()  # ignore `self` and `form_data` parameters
    existing_params = [v for v in dest_signature.parameters.values() if v.kind != inspect.Parameter.VAR_KEYWORD]
    dest_function.__signature__ = dest_signature.replace(parameters=existing_params + params_to_copy)
    return dest_function


FormT = TypeVar("FormT", bound=Form)


class FormController(Controller):
    """
    Base class allowing to write controllers to handle forms

    Example:
    ```
    class MyClass(FormController, MyJinjaController, router=router):
        url = "/my_form"
        template = "form.html"
        form = MyForm

        def on_valid_form(self, form: MyForm, value = Depends(some_other_dependency)):
            return "OK"
    ```
    """

    url: ClassVar[str]
    template: ClassVar[str]
    form: ClassVar[Type[FormT]]

    name: str = None
    "Can be used to name the GET route, for usage in url_for"

    def __init_subclass__(cls, router: APIRouter = None, **kwargs):
        if router is not None:

            @router.get(cls.url, name=cls.name)
            def main_page(self):
                # getting the context must be the first action done here: it allows the
                # method to raise AccessDenied if needed.
                self.context = self.get_context()
                self.context.update(form=cls.form())
                return self.render(cls.template, context=self.context)

            @router.post(cls.url)
            def handle_form_page(self, form_data=WholeForm(), **kw):
                # getting the context must be the first action done here: it allows the
                # method to raise AccessDenied if needed.
                self.context = self.get_context()
                form = cls.form(formdata=form_data)
                self.context.update(form=form)
                if not form.validate():
                    return self.on_invalid_form(form, **kw)
                else:
                    return self.on_valid_form(form, **kw)

            cls.main_page = main_page
            cls.handle_form_page = _copy_dependencies(cls.on_valid_form, handle_form_page)

        super().__init_subclass__(router=router, **kwargs)

    def on_valid_form(self, *args, **kwargs):
        """
        Override this method to handle the form data
        """

    def on_invalid_form(self, *args, **kwargs):
        """
        Override this method to return a different response in case of an invalid form.
        The default implementation should be enough for most cases.
        """
        return self.render(self.template, context=self.context, status_code=400)

    def get_context(self):
        """
        Override this to provide extra context to the jinja template.
        """
        return {}



from dataclasses import dataclass
from typing import Literal, Type, TypeVar, get_type_hints

from pydantic import BaseSettings


class ConfigError(Exception):
    pass


class AuthConfig(BaseSettings):
    url: str
    cookie_domain: str
    pubkey_path: str


class DatabaseConfig(BaseSettings):
    url: str


class LoggingConfig(BaseSettings):
    format: Literal["console", "json"] = "console"


class RabbitMQConfig(BaseSettings):
    url: str


class CORSConfig(BaseSettings):
    allowed_origin_re: str = ""


AnyConfig = TypeVar("AnyConfig", bound=BaseSettings)
ChloreConfig = TypeVar("ChloreConfig", AuthConfig, CORSConfig, DatabaseConfig, LoggingConfig, RabbitMQConfig)


@dataclass
class _InternalConfig:
    auth: AuthConfig = None
    cors: CORSConfig = None
    logging: LoggingConfig = None
    database: DatabaseConfig = None
    rabbitmq: RabbitMQConfig = None

    def wants(self, type: Type[AnyConfig]) -> bool:
        return type in ChloreConfig.__constraints__

    def register(self, type: Type[ChloreConfig], value: ChloreConfig):
        table = {type: field_name for field_name, type in get_type_hints(_InternalConfig).items()}
        setattr(self, table[type], value)


CONFIG = _InternalConfig()


def from_env(t: Type[AnyConfig], with_prefix: str = "") -> AnyConfig:
    class WithPrefix(t):
        class Config:
            env_prefix = with_prefix

    result = WithPrefix()
    if CONFIG.wants(t):
        CONFIG.register(t, result)
    return result
